"""
----------------------------------------------------------------------------------------------------
Written by:
  - Yovany Dominico Girón (y.dominico.giron@elprat.cat)

for Ajuntament del Prat de Llobregat
----------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------
"""
from abc import ABC, abstractmethod
from typing import TypeVar, Protocol, Union

from pydantic import BaseModel

D = TypeVar('D', bound=object)  # Domain class representation
S = TypeVar('S', bound=BaseModel)  # Schema class representation


class SchemaMapper(Protocol[D, S]):
    """SchemaMapper class"""

    @staticmethod
    @abstractmethod
    def map_from_schema_to_domain(schema: Union[S, dict]) -> D:
        """Map to Domain from Schema"""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def map_from_domain_to_schema(domain: D) -> S:
        """Map to Schema from Domain"""
        raise NotImplementedError


class NomenclatorMapper(ABC):
    """Classe Nomenclator Mapper"""

    DomainClass: type
    SchemaClass: type

    @classmethod
    @abstractmethod
    # type: ignore
    def map_from_schema_to_domain(cls, schema: Union[S, dict]) -> D:
        """Map to Domain from Schema"""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def map_from_domain_to_schema(cls, domain: D) -> S:  # type: ignore
        """Map to Schema from Domain"""
        raise NotImplementedError


class BaseNomenclatorMapper(NomenclatorMapper):
    """BaseNomenclator Mapper class"""

    @classmethod
    # type: ignore
    def map_from_schema_to_domain(cls, schema: Union[S, dict]) -> D:
        """Map to Domain from Schema"""

        if not cls.DomainClass:
            raise NotImplementedError(
                "The domain class (DomainClass) must be defined in the subclass")

        if not isinstance(schema, dict):
            schema = schema.model_dump()

        schema["identifier"] = schema.pop("id", None)
        return cls.DomainClass(**schema)

    @classmethod
    def map_from_domain_to_schema(cls, domain: D) -> S:  # type: ignore
        """Map to Schema from Domain"""

        if not cls.SchemaClass:
            raise NotImplementedError(
                "The persist class (SchemaClass) must be defined in the subclass")

        return cls.SchemaClass(**domain.__dict__)
