"""
----------------------------------------------------------------------------------------------------
Written by:
  - Yovany Dominico Girón (y.dominico.giron@elprat.cat)

for Ajuntament del Prat de Llobregat
----------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------
"""
from datetime import datetime, date, time
from typing import Any, Optional, Dict, Union, Literal, List, Tuple

from pydantic import BaseModel, RootModel, Field

from nomenclators_archetype.domain.repository.builders import COMPARISON_OPERATIONS, LITERAL_OPERATIONS


class SchemaPagination(BaseModel):
    """Schema Pagination class"""

    page: int = Field(
        ge=0, description="Número de página (en caso de utilizarse, debe ser >= 0)")
    element_for_page: int = Field(
        gt=0, description="Elementos por página (en caso de utilizarse, debe ser > 0)")
    limit: int = Field(
        gt=0, description="Máximo total de elementos a analizar (en caso de utilizarse, debe ser > 0)")
    element_sort: str = Field(...,
                              description="Campo por el cual se desea ordenar")
    sort: Literal["ASC", "DESC"] = Field(
        "DESC", description="Orden 'ASC' para ordenar Ascendentemente o 'DESC' para ordenar Descendentemente")

    def sort_mapped(self) -> List[Tuple[str, str]]:
        """Get sorte mapped"""
        return [(self.element_sort, self.sort)]


class SchemaFilterValue(BaseModel):
    """Schema FilterValue class: allows you to assign a direct value, a like/not like or a range"""

    eq: Optional[Union[str, int, float, datetime, date, time]] = None
    noteq: Optional[Union[str, int, float, datetime, date, time]] = None
    like: Optional[str] = None
    notlike: Optional[str] = None
    inc: Optional[list] = None
    notinc: Optional[list] = None
    gt: Optional[Union[int, float, datetime, date, time]] = None
    gte: Optional[Union[int, float, datetime, date, time]] = None
    lt: Optional[Union[int, float, datetime, date, time]] = None
    lte: Optional[Union[int, float, datetime, date, time]] = None


class SchemaFilters(RootModel[Dict[str, Union[SchemaFilterValue, list, str, int, float, datetime, date, time]]]):
    """Schema Filters class: allows you to assign a filter to a field"""

    def mapped(self) -> Dict[str, Any]:
        """Get filters mapped"""

        filters = {}
        for key, value in self.root.items():

            if isinstance(value, list):
                for item in list(value):
                    if isinstance(item, dict):
                        self._update_filter_value(filters, key, item)
                    elif isinstance(item, SchemaFilterValue):
                        self._update_filter_value(
                            filters, key, item.model_dump())
                    else:
                        raise ValueError(
                            f"Invalid filter value for key {key}: {item}")
            elif isinstance(value, SchemaFilterValue):
                self._update_filter_value(filters, key, value.model_dump())
            else:
                filters[key] = value

        return filters

    def _update_filter_value(self, filters: dict, key: str, value: dict) -> None:
        """Update schema filter value"""
        for sub_key, sub_value in value.items():
            if sub_value is not None:
                filters.update(self._map_filter(
                    key, sub_key, sub_value))

    def _map_filter(self, key: str, comparison_key: str, comparison_value: Any) -> dict:
        """
        Apply map filter to the key definition

        :param key: Key to filter
        :param comparison_key: comparing operator
        :param comparison_value: Value to comparison
        """
        if (
            (comparison_key.upper() in COMPARISON_OPERATIONS and isinstance(comparison_value, (str, int, float, datetime, date, time))) or
            (comparison_key.upper() in LITERAL_OPERATIONS and isinstance(
                comparison_value, (list, str)))
        ):
            return {f"{key}_{comparison_key.lower()}": comparison_value}
        else:
            return {}


class SchemaQueryParam(BaseModel):
    """Schema QueryParam class: allows you to assign a pagination and filter"""

    pagination: Optional[SchemaPagination] = None
    filter: Optional[SchemaFilters] = None
